/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: CustomNew.cpp,v 1.1 2003/01/26 10:36:53 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/


#include <PalmOS.h>
#include "Device.h"


/**
 * @file CustomNew.cpp
 * Overloaded global operators new and delete. The default operators
 * throw a std:bad_alloc exception, which requires C++ exception
 * handling to be enabled. The customized versions will invoke
 * Device::panic in case of insufficient memory.
 */


/**
 * Operator new that panics instead of throwing std::bad_alloc.
 */
void * operator new (UInt32 size)
{
	// C++ specs require the allocation of at least one byte, regardless of request size
	if (size == 0L)
	{
		size = 1L;
	}

    void *p = MemPtrNew(size);
    
    if (p == NULL)
    {
    	Device::panic("Not enough memory for operator new");
    }

	return p;    	    	
}


/**
 * Operator new that panics instead of throwing std::bad_alloc.
 */
void * operator new[] (UInt32 size)
{
	// C++ specs require the allocation of at least one byte, regardless of request size
	if (size == 0L)
	{
		size = 1L;
	}

    void *p = MemPtrNew(size);
    
    if (p == NULL)
    {
    	Device::panic("Not enough memory for operator new");
    }

	return p;    	    	
}


/**
 * Operator delete that is compatible with the customized operators new.
 */
void operator delete(void* p)
{
	Err error = MemPtrFree(p);
	
	// It's probably a good idea to stop a program which has corrupted memory
	ErrFatalDisplayIf(error != errNone, "Invalid pointer for delete.");
}


/**
 * Operator delete that is compatible with the customized operators new.
 */
void operator delete[] (void* p)
{
	Err error = MemPtrFree(p);
	
	// It's probably a good idea to stop a program which has corrupted memory
	ErrFatalDisplayIf(error != errNone, "Invalid pointer for delete.");
}
