/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Starter.cpp,v 1.3 2003/05/31 20:05:28 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

/**
 * TODO: Figure out proper way to unpause after an alarm dialog 
 */

#include "./HostControl.h"
#include <PalmOS.h>

#include "Customization.h"
#include "../../Rsc/RazorRsc.h"

#include "Device.h"
#include "app/Presentation.h"


/**
 * @file Starter.cpp
 * The main application. This is the bare skeleton of a functional
 * PalmOS application. It delegates most of the work to a Presentation
 * object.
 */



// ***********************************************************************

Presentation *presentation;

#ifdef MODE_PROFILING
static UInt32                  periodCounter;      // The current period
#endif

// ***********************************************************************



/**
 * Perform the specified menu command
 *
 * @param command menu item id
 *
 * @return true if the command has been handled
 */
static Boolean MainFormDoCommand(UInt16 command)
{
	Boolean handled = false;

	switch (command)
	{
		case MainOptionsAboutRazor:
			presentation->pause();

			MenuEraseStatus(0);					// Clear the menu status from the display.
			FormType* frmP = FrmInitForm(AboutForm);
			FrmDoDialog (frmP);					// Display the About Box.
			FrmDeleteForm (frmP);
			handled = true;
			
			presentation->resume();
		break;
	}
	
	return handled;
}


/**
 * The event handler for the "MainForm" of this application.
 * 
 * @param eventP a pointer to an EventType structure
 *
 * @return true if the event has been handled and should not be passed to a higher level handler.
 */
static Boolean MainFormHandleEvent(EventType *eventP)
{
	Boolean handled = false;

	switch (eventP->eType) 
	{
		case menuEvent:
			return MainFormDoCommand(eventP->data.menu.itemID);

		case frmOpenEvent:
			{   // This scope is required for PRC-Tools 2.0
				FormType *frmP = FrmGetActiveForm();
				FrmDrawForm(frmP);
			}
			// Begin the presentation
			presentation->begin();

			handled = true;			
			break;

		default:
			break;		
	}
	
	return handled;
}




/**
 * Load form resources and set the event handler for the form loaded.
 *
 * @param eventP a pointer to an EventType structure
 *
 * @return true if the event has been handled and should not be passed to a higher level handler.
 */
static Boolean AppHandleEvent(EventType *eventP)
{
	if (eventP->eType == frmLoadEvent)
	{
		// Load the form resource.
		UInt16 formId = eventP->data.frmLoad.formID;
		FormType *frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time it receives an
		// event.
		switch (formId)
		{
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;

			default:
				ErrNonFatalDisplay("Invalid Form Load Event");
				break;

		}
		return true;
	}
	
	return false;
}



/**
 * Handle keydown events where the key is a command vchar.
 * 
 * @return true if the event should not be passed to Palm OS event handling.
 */
static Boolean handleCommandChar(EventType *eventP)
{
	// Treat virtual keys properly. Some can safely be ignored, 
	// others require the display to be released, so they can show
	// their dialog boxes.
	WChar chr = eventP->data.keyDown.chr;

	// Swallow events notifying us of hard key presses.  We poll instead.
	// TODO: What about page up/down? Ignore them too?
	if (chr >= vchrHard1 && chr <= vchrHard4 &&
		!(eventP->data.keyDown.modifiers & poweredOnKeyMask))
	{
		return true;
	}

	
	// Ignore these
	// TODO: Should Tsm# really be swallowed? This might prevent text input on Japanese devices?
	else if (
		chr == vchrCommand || chr == vchrFind || chr == vchrCalc ||
		(chr >= vchrTsm1 && chr <= vchrTsm4) || chr == vchrKeyboard || (chr >= vchrRonamatic && chr <= vchrKeyboardNumeric)
	)
	{
		return true;
	} 

    // Alarms and the attention manager require special treatment.
	//       On OS 4, datebook alarms send vchrAttnStateChanged when dialog opens, and send
	//       winExitEvent, winEnterEvent, vchrAttnStateChanged and vchrAlarm when dialog is dismissed.
	//       On OS 3.5, vchrAlarm is sent when dialog opens, and winEnterEvent when dialog is dismissed.
    else if (chr == vchrAttnStateChanged)
    {
        // The attention manager announces "change of state" through this key.
        // TODO: Is is possible to figure out what the state is then?
        presentation->pause();
    }

    else if (chr == vchrAlarm)
    {
        if (!Device::supports40)
        {
            // On OS before 4.0 the vchrAlarm is sent if the alarm dialog pops up.
    		HostTraceOutput2("Starter: pause b/c vchr %hu %hu", chr, eventP->data.keyDown.modifiers);
    		presentation->pause();
        }
        else
        {
            // On OS 4.0+ the vchrAlarm is sent if the alarm dialog is dismissed.
            presentation->resume();
        }
    }

    else if (chr == vchrMenu)
    {
        if (presentation->enableMenu())
        {
            HostTraceOutput("Starter: pause b/c vchrMenu");            
            presentation->pause();
        }
        else
        {
            HostTraceOutput("Starter: No pause for menu");
            return true;
        }
    }
        
	// Release display for these
	else if (
		chr == vchrLock || chr == vchrBrightness || chr == vchrContrast ||
		chr == vchrHardBrightness || chr == vchrHardContrast
	)
	{
		HostTraceOutput2("Starter: pause b/c vchr %hu %hu", chr, eventP->data.keyDown.modifiers);

		presentation->pause();
	}
	
	// Special treatment for vchrLowBattery
	else if (chr == vchrLowBattery)
	{
		// vchrLowBattery is posted occasionally, even though
		// the battery is not low, and no warning dialog will be displayed.
		// This behavior has been seen on OS 3.3 devices.
		HostTraceOutputTL(appErrorClass, "Starter: received vchrLowBattery");

		if (Device::supports35)
		{
			// No wrong vchrLowBattery events known on OS 3.5+.
			presentation->pause();
		}
		else
		{
			// Is this really a low battery situation?
			UInt16 warnThreshold;
			UInt16 currentVoltage = SysBatteryInfo(false, &warnThreshold, NULL, NULL, NULL, NULL, NULL);
			
			if (currentVoltage < warnThreshold)
			{
				presentation->pause();   // Pause for upcoming warning dialog
			}
			else
			{
				return true;   // False alert! Continue...
			}
		}
	}

	return false;
}



/**
 * The event loop for the application.  
 */
static void AppEventLoop()
{
	UInt16 error;
	EventType event;

	// The game has not yet begun. Let PalmOS handle all events until
	// frmOpenEvent has been received on the Main form.
	do
	{
		EvtGetEvent (&event, evtWaitForever);
		if (! SysHandleEvent (&event))
			if (! MenuHandleEvent (NULL, &event, &error))
				if (! AppHandleEvent (&event))
					FrmDispatchEvent (&event);

	} while ((event.eType != appStopEvent) && (!(presentation->hasBegun())));

#ifdef MODE_PROFILING
	periodCounter = 0;
#endif		


	// The regular event handling loop
	do
	{
		// Wait until the next game period.
		EvtGetEvent (&event, presentation->getTimeUntilNextPeriod());
		
#ifndef MODE_PROFILING
		switch(event.eType)
		{
			case nilEvent:
				// This usally means that the next period has been reached.
				break;

			case winEnterEvent:
				// Detect entering the game's window.  Resume drawing to our window.
				if (event.eType == winEnterEvent)
				{
					HostTraceOutput("Starter: winEnterEvent");
		
					// In the current code, the menu doesn't remove itself when it receives
					// a winExitEvent.
					if (event.data.winEnter.enterWindow == (WinHandle) FrmGetFormPtr(MainForm) &&
						event.data.winEnter.enterWindow == (WinHandle) FrmGetFirstForm ())
					{
						// Sometimes we can enter the game's window without knowing it was 
						// ever left.  In that case the pause time will not have been recorded.
						// Presentation is smart enough to handle that, and will set the current 
						// period back to it's beginning.
						presentation->resume();
					}
				}
				break;

			case winExitEvent:
				// Detect exiting the game's window.  This must be checked for.  At this
				// point there probably exists another window which may cover part of
				// the MainView window.  Suppress drawing.  Otherwise drawing may draw
				// to part of the window covered by the new window.
				HostTraceOutput("Starter: winExitEvent");
	
				if (event.data.winExit.exitWindow == (WinHandle) FrmGetFormPtr(MainForm))
				{
					presentation->pause();
				}

				break;

			case menuCmdBarOpenEvent:
				continue;
				break;

			case menuOpenEvent:
				continue;
				break;

			case keyDownEvent:
				HostTraceOutput1("Starter: keyDownEvent %hu", event.data.keyDown.chr);

				if (event.data.keyDown.modifiers & commandKeyMask)
				{
					// Handle the command vchar
					if (handleCommandChar(&event))
					{
						continue;
					}
				}
				else
				{
					// Regular keys are passed through to the application.
					presentation->receiveEvent(&event);
				}
				break;

			default:				
				HostTraceOutput1("Starter: event %hu", event.eType);
				presentation->receiveEvent(&event);
		}
#endif  // MODE_PROFILING
		
		
		// If it's time, go to the next time period
		if (presentation->getTimeUntilNextPeriod() == 0)
		{		
#ifdef MODE_PROFILING
			periodCounter++;
			if (periodCounter == profilingBegin)
			{
				// Turn on profiling
				HostProfileInit(profilingMaxCalls, profilingMaxDepth);
				HostProfileStart();
			}
			else if (periodCounter == profilingEnd)
			{
				// Turn off profiling
				HostProfileStop();
				HostProfileDump(profilingFile);
				HostProfileCleanup();

				// Emit sound signal to alert the developer
				SndCommandType soundCommand;
				soundCommand.cmd = sndCmdFreqDurationAmp;
				soundCommand.param1 = 440;
				soundCommand.param2 = 1000;
				soundCommand.param3 = sndMaxAmp;
				SndDoCmd(NULL, &soundCommand, 0);
			}
#endif  // MODE_PROFILING

			presentation->nextPeriod();
		}




		// Regular event handling by the OS.
		if (!SysHandleEvent(&event))
		{
#ifndef MODE_PROFILING
			if (!MenuHandleEvent(0, &event, &error))
			{
				if (!AppHandleEvent(&event))
				{
					FrmDispatchEvent(&event);
				}
			}
#endif  // MODE_PROFILING			
		}

	} while (event.eType != appStopEvent);
}


/**
 * Prepare the App for running
 */
static Err AppStart(UInt16 launchFlags)
{
#if ERROR_CHECK_LEVEL == ERROR_CHECK_FULL
// This slows the device down to a crawl :-(
//	MemSetDebugMode(memDebugModeCheckOnChange | memDebugModeScrambleOnChange | memDebugModeFillFree);
#endif
    presentation = NULL;

	HostTraceInit();
	
	Device::init();

	Err error = Device::romVersionCompatible(appMinRomVersion, appMaxRomVersion, launchFlags);
	if (error != errNone) 
		return error;

	// Establish the presentation
	presentation = new Presentation();

	return errNone;
}


/**
 * Shutdown the App (save prefs, restore display to normal)
 */
static void AppStop()
{
	// Close all the open forms.
	FrmCloseAllForms();

    if (presentation != NULL)
    {
	    delete presentation;
	}

	Device::release();

	HostTraceClose();
}


/**
 * This is the main entry point for the application.
 *
 * @param cmd word value specifying the launch code. 
 * @param cmdPB pointer to a structure that is associated with the launch code. 
 * @param launchFlags word value providing extra information about the launch.
 *
 * @return Result of launch
 */
static UInt32 StarterPalmMain(UInt16 cmd, void */*cmdPBP*/, UInt16 launchFlags)
{
	if (cmd == sysAppLaunchCmdNormalLaunch)
	{
        // The execution of the application is enclosed by the OS try/catch mechanism.
        // This allows exiting the application from anywhere (e.g. Device::panic).
        ErrTry
        {
    		Err error = AppStart(launchFlags);
    		if (error != errNone)
    			return error;
    				
    		FrmGotoForm(MainForm);
    		AppEventLoop();
        }
        ErrCatch(inErr)
        {
            // Application failure. Cleanup work will be done by AppStop later.
        } ErrEndCatch
        
		AppStop();

		return errNone;
	}
	else
	{
		return errNone;
	}
}


/**
 * This is the public main entry point for the application.
 *
 * @param cmd word value specifying the launch code. 
 * @param cmdPB pointer to a structure that is associated with the launch code. 
 * @param launchFlags word value providing extra information about the launch.
 *
 * @return Result of launch
 */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	return StarterPalmMain(cmd, cmdPBP, launchFlags);
}

