/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: BlitterSupport.cpp,v 1.2 2003/05/29 15:45:24 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

#include "./HostControl.h"
#include <PalmOS.h>
#include <BmpGlue.h>
#include "Customization.h"
#include "Device.h"
#include "gfx/Display.h"
#include "gfx/Canvas.h"
#include "gfx/BlitterSupport.h"

using namespace prvrazor;



Err BlitterSupport::copyBitmap(WinHandle sourceWindow, UInt16 ftrId, Coord width, Coord height, BitmapType **bitmapType, void **bitmapBits)
{
	// TODO: Shall treatment for indirect bitmaps be provided?

	// Determine req'd memory size	
	UInt32 dataSize;
	UInt32 headerSize;
	BitmapType *header;
	
	if (!Device::supports35)
	{
		dataSize = ((width * height) * Display::depth) / 8;
		headerSize = 0;
		header = NULL;
	}
	else if (!Device::supports40)
	{
		header = WinGetBitmap(sourceWindow);
		dataSize = (UInt32)BmpBitsSize(header);
		headerSize = ((UInt32)BmpSize(header)) - dataSize;		
	}	
	else
	{
		header = WinGetBitmap(sourceWindow);
		BmpGetSizes(header, &dataSize, &headerSize);
	}

	// Ignore header, if caller is not interested in its address.
	if (bitmapType == NULL)
	{
		headerSize = 0;
	}

	UInt32 totalSize = dataSize + headerSize;
	
	void *ftrMemory;
	Err error = FtrPtrNew (Device::creatorID, ftrId, totalSize, &ftrMemory);
	if (error)
		Device::panic("Couldn't allocate feature memory.");		

	if (headerSize > 0)
	{
		DmWrite(ftrMemory, 0, (void *)header, headerSize);
		*bitmapType = (BitmapType *)ftrMemory;
	}
	else if (bitmapType != NULL)
	{
		*bitmapType = NULL;
	}

	DmWrite(ftrMemory, headerSize, BlitterSupport::getWindowBits(sourceWindow), dataSize);
	*bitmapBits = (void *)(((UInt8 *)ftrMemory) + headerSize);
		
	return errNone;	
}


BitmapType *BlitterSupport::createIndirectV2Copy(const BitmapType *bitmapP, UInt32 desiredDepth)
{
	const BitmapType *candidateP = BlitterSupport::getHighDensityMember(bitmapP, desiredDepth);
	ErrNonFatalDisplayIf(candidateP == NULL, "Found no matching bitmap family member.");

	const RscBitmapTypeV3 *candidateV3P = (RscBitmapTypeV3 *)candidateP;

	// Further process the candidate bitmap
	Coord width, height;
	BmpGlueGetDimensions(candidateP, &width, &height, NULL);

	const Int16 pixelSize = BlitterSupport::getPixelSize(candidateP);
	const Boolean isCompressed = BlitterSupport::isCompressed(candidateP);

	// BmpGetBits would return incorrect results...
	const UInt8 *bits = ((UInt8*)candidateV3P) + sizeof(RscBitmapTypeV3);

	// Compensate for differences between V2 and V3.
	if (isCompressed)
	{
		bits += 2;
	}
	
	RscBitmapTypeV2 * const result = BlitterSupport::createBitmapTypeV2(width, height, pixelSize, bits);

	// Set the appropriate flags.
	if (isCompressed)
	{
		result->flags.compressed = 1;

		// Find the compression type and then set it in the copy.
		result->compressionType = ((RscBitmapTypeV2 *)(bitmapP))->compressionType;
	}

	return (BitmapType *)result;
}


Int16 BlitterSupport::getPixelSize(const BitmapType *bitmapP)
{
	return (bitmapP->pixelSize == 0) ? 1 : bitmapP->pixelSize;
}


Boolean BlitterSupport::isCompressed(const BitmapType *bitmapP)
{
	return bitmapP->flags.compressed;
}


void *BlitterSupport::getWindowBits(WinHandle window)
{
	if (Device::supports35)
	{
		return BmpGetBits(WinGetBitmap(window));
	}
	else
	{
		return (void *)(window->displayAddrV20);
	}
}


BitmapType *BlitterSupport::getWindowBitmap(WinHandle window)
{
	if (Device::supports35)
	{
		return WinGetBitmap(window);
	}
	else
	{
		return (BitmapType *)(window->bitmapP);
	}
}


RscBitmapTypeV2 *BlitterSupport::createBitmapTypeV2
(
		Int16 width, Int16 height,
		Int16 pixelSize, 
		const UInt8 *bits
)
{
	Int16 rowBytes = ((((UInt32)width * pixelSize) + 15) >> 4) << 1;
	Int32 totalSize = sizeof(RscBitmapTypeV2) + sizeof(UInt8 *);

	RscBitmapTypeV2 *bmp = (RscBitmapTypeV2 *)MemPtrNew(totalSize);
	if (bmp == NULL)
	{
		Device::panic("Not enough memory for Bitmap V2 structure");
	}

	MemSet(bmp, sizeof(RscBitmapTypeV2), 0);

	bmp->width = width;
	bmp->height = height;
	bmp->rowBytes = rowBytes;
	bmp->pixelSize = pixelSize;
	bmp->version = 2;

	// Make this an indirect bitmap, which points to the bits which were passed into this method.		
	UInt8 *ptr = (UInt8*)(bmp + 1);
	*(const UInt8**)ptr = bits;
	bmp->flags.indirect = true;

#if ERROR_CHECK_LEVEL == ERROR_CHECK_FULL
	HostTraceOutput1("Structure size %ld", totalSize);
	HostTraceOutput1("Row bytes %hd", rowBytes);
	HostTraceOutput1("Pixelsize %hd", pixelSize);
#endif
		
	return bmp;
}


const BitmapType *BlitterSupport::getHighDensityMember(const BitmapType *bitmapP, UInt32 desiredDepth)
{
	// Traverse through bitmap family to appropriate family member
	// Needs to be version 3, density 144, desiredDepth
	Boolean found = false;
	const BitmapType *candidateP = bitmapP;
	const RscBitmapTypeV3 *candidateV3P;
	do
	{
		Int16 pixelSize = BlitterSupport::getPixelSize(candidateP);
		UInt8 version = candidateP->version;

		HostTraceOutput2("HD cand. pxsize %hd, version %hd", pixelSize, version);
		
		if ((pixelSize == desiredDepth) && (version == 3))
		{
			candidateV3P = (RscBitmapTypeV3 *)candidateP;
			HostTraceOutput1("HD cand. density %hd", candidateV3P->density);
			if (candidateV3P->density == 144)
			{
				found = true;
				break;
			}
		}

		// Search next family member
		// Needs user written code, b/c this needs support for high density, but is not running on OS5
		switch(version)
		{
			case 1:
				{
					// Skip dummy bitmap
					if (BlitterSupport::getPixelSize(candidateP) == 0xFF)
					{
						candidateP = (const BitmapType *)(((const RscBitmapTypeV1 *)candidateP) + 1);
						continue;
					}
				}
			case 2:
				{
					const UInt16 nextDepthOffset = ((RscBitmapTypeV2 *)candidateP)->nextDepthOffset;
					if (nextDepthOffset != 0)
					{
						candidateP = (const BitmapType *)((UInt32*)candidateP + nextDepthOffset);
					}
					else
					{
						candidateP = NULL;
					}
				}
				break;
			case 3:
				{
					const UInt32 nextBitmapOffset = ((RscBitmapTypeV3 *)candidateP)->nextBitmapOffset;
					if (nextBitmapOffset != 0)
					{
						candidateP = (const BitmapType *)((UInt8*)candidateP + nextBitmapOffset);
					}
					else
					{
						candidateP = NULL;
					}
				}
				break;
		}		
	} 
	while(candidateP != NULL);

	if (!found)
	{
		return NULL;
	}
	else
	{
		return candidateP;
	}
}
