/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Direct8bppBlitter.cpp,v 1.3 2003/03/17 22:29:20 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

#include "Customization.h"
#include <BmpGlue.h>
#include "Device.h"
#include "gfx/Direct8bppBlitter.h"
#include "gfx/Canvas.h"
#include "gfx/BlitterSupport.h"

using namespace prvrazor;

/*
 * TODO:
 * - Convert all routines to ASM for CW and GCC.
 * - Which of the fralloc (+) and movem.l xxx, (-SP) stuff is really neccessary?
 */


Direct8bppTransparentBitmapArray::Direct8bppTransparentBitmapArray(const Canvas *canvas, UInt16 numBitmaps, UInt16 bitmapID, UInt16 maskID)
	: BitmapArray(numBitmaps), bitmaps(new Direct8bppTransparentBitmap[numBitmaps]),
	bitmapFtrFirstNum(bitmapID), maskFtrFirstNum(maskID)
{
	Err error = errNone;

	// Determine bitmap extents
	{
		const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
		ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
		BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
	
		BmpGlueGetDimensions(resP, &this->width, &this->height, NULL);
	
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
	}


	// Prepare all bitmaps
	RectangleType bitmapRect;
	bitmapRect.topLeft.x = bitmapRect.topLeft.y = 0;
	bitmapRect.extent.x = width; bitmapRect.extent.y = height;

	const WinHandle image = WinCreateOffscreenWindow(width, height, screenFormat, &error);
	const WinHandle mask = WinCreateOffscreenWindow(width, height, screenFormat, &error);

	WinSetDrawWindow(mask);
	WinSetForeColor(0);
	WinSetBackColor(255);
	
	for(UInt16 i = 0; i < numBitmaps; i++)
	{
		Direct8bppTransparentBitmap *const bitmap = const_cast<Direct8bppTransparentBitmap *>(&bitmaps[i]);
		// Paint sprite-mask to offscreen window
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, maskID + i);
			ErrNonFatalDisplayIf( !resH, "Missing mask" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		
			WinSetDrawWindow(mask);
			WinDrawBitmap (resP, 0,0);
			error = BlitterSupport::copyBitmap(mask, maskID + i, width, height, static_cast<BitmapType **>(NULL), reinterpret_cast<void **>(&(bitmap->maskBits)));
	
			MemPtrUnlock(resP);
			DmReleaseResource( resH );				
		}

		// Paint sprite-bitmap to offscreen window
		{
			const MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID + i);
			ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
			BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
	
			WinSetDrawWindow(image);
			WinDrawBitmap (resP, 0,0);
			// TODO: If I circumvented the OS and used the maskBits, I wouldn't need two offscreen windows.
			WinCopyRectangle(mask, image, &bitmapRect, 0, 0, winMask);
			
			error = BlitterSupport::copyBitmap(image, bitmapID + i, width, height, static_cast<BitmapType **>(NULL), reinterpret_cast<void **>(&(bitmap->bitmapBits)));
		
			MemPtrUnlock(resP);
			DmReleaseResource( resH );
		}	
	}

	WinSetDrawWindow(canvas->drawWindow);
	WinDeleteWindow(image, false);
	WinDeleteWindow(mask, false);
}



Direct8bppTransparentBitmapArray::~Direct8bppTransparentBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		FtrPtrFree(Device::creatorID, bitmapFtrFirstNum + i);
		FtrPtrFree(Device::creatorID, maskFtrFirstNum + i);
	}

	delete [] bitmaps;
}



void Direct8bppTransparentBitmapArray::draw(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	const Direct8bppTransparentBitmap *const bitmap = &bitmaps[index];

	// Determine memory address of bits
	const UInt8 *foregroundBits = bitmap->bitmapBits;
	const UInt8 *backgroundBits = bitmap->maskBits;
	UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Clipping
	Int16 xOffset = 0;
	Int16 xExtent = width;
	if (x < 0)
	{
		xOffset = -x;
		xExtent -= xOffset;
	}
	
	if (x + width > canvas->width)
	{
		xExtent -= ((x + width) - canvas->width);
	}

	if (xExtent <= 0)
	{
		return;
	}
	

	Int16 yOffset = 0;
	Int16 yExtent = height;
	if (y < 0)
	{
		yOffset = 0 - y;
		yExtent -= yOffset;
	}
	
	if (y + height > canvas->height)
	{
		yExtent -= ((y + height) - canvas->height);
	}

	if (yExtent <= 0)
	{
		return;
	}
	

	// Paint
	UInt16 sourceOdd;
	{
		UInt16 sourceOffset = (width * yOffset + xOffset);
		foregroundBits += sourceOffset;
		backgroundBits += sourceOffset;	
		
		sourceOdd = sourceOffset & 1;
	}

	// TODO: Should this rather be canvas->cols?	
	drawBits += (canvas->width * (y + yOffset) + x + xOffset);

	const UInt16 destOdd = (x + xOffset) & 1;

	UInt16 colQuads = (xExtent - sourceOdd) >> 3;
	UInt16 colRemainder = (xExtent - sourceOdd) & 7;

	if (!sourceOdd && !destOdd && !colRemainder)
	{
		// Everything is perfect for a blindingly fast blit :-)
#ifdef MODE_M68K

#ifdef __GNUC__
/*
	%0 foregroundBits
	%1 backgroundBits
	%2 drawBits, 
	%3 (yExtent - 1)
	%4 (colQuads - 1),
	%5 (width - xExtent)
	%6 canvas->width - xExtent
*/

	yExtent--;
	const UInt32 canvasRemainder = canvas->width - xExtent;
	const UInt32 remainWidth = width - xExtent;
	colQuads--;

	asm("
rowloop:
	move.w %4, %%d2;

colloop:
	move.l (%2), %%d3;
	move.l (%1)+, %%d7;
	and.l  %%d7, %%d3;
	or.l   (%0)+, %%d3;
	move.l %%d3, (%2)+;	

	move.l (%2), %%d3;
	move.l (%1)+, %%d7;
	and.l  %%d7, %%d3;
	or.l   (%0)+, %%d3;
	move.l %%d3, (%2)+;	

dbra.w  %%d2, colloop;

	add.l 	%5, %0;
	add.l	%5, %1;
	add.l   %6, %2;
	
dbra.w  %3, rowloop;
	" : 
	  : "a" (foregroundBits), "a" (backgroundBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (remainWidth), "r" (canvasRemainder)
	  : "%d2", "%d3", "%d7");
#else
		drawM68KAsm(
			foregroundBits, backgroundBits, drawBits, 
			yExtent, colQuads,
			width, canvas->width, xExtent);		
#endif  // ifdef __GNUC__
#else   // ! MODE_M68K
		for (UInt16 row = 0; row < yExtent; row++)
		{
			for (UInt16 col = 0; col < colQuads; col++)
			{
				UInt32 pixelByte = (*((UInt32 *)drawBits));
				pixelByte &= (*((UInt32 *)backgroundBits));
				pixelByte |= (*((UInt32 *)foregroundBits));
				(*((UInt32 *)drawBits)) = pixelByte;
		
				foregroundBits += 4;
				backgroundBits += 4;
				drawBits += 4;

				pixelByte = (*((UInt32 *)drawBits));
				pixelByte &= (*((UInt32 *)backgroundBits));
				pixelByte |= (*((UInt32 *)foregroundBits));
				(*((UInt32 *)drawBits)) = pixelByte;
		
				foregroundBits += 4;
				backgroundBits += 4;
				drawBits += 4;
			}
		
			foregroundBits += width - xExtent;
			backgroundBits += width - xExtent;
			drawBits += canvas->width - xExtent;
		}
#endif  // MODE_M68K
	}
	else
	{
		// Welcome to a world of lameness :-(
		for (Int16 row = 0; row < yExtent; row++)
		{
			for (Int16 col = 0; col < xExtent; col++)
			{
				if (!(*backgroundBits))
				{								
					(*drawBits) = (*foregroundBits);
				} 
				foregroundBits++;
				backgroundBits++;
				drawBits++;
			}
	
			foregroundBits += width - xExtent;
			backgroundBits += width - xExtent;
			drawBits += canvas->width - xExtent;
		}
	}
}



#ifdef __MWERKS__
asm void Direct8bppTransparentBitmapArray::drawM68KAsm(
	const void *foregroundBits, const void *backgroundBits, void *drawBits, 
	Int16 yExtent, UInt16 colQuads,
	UInt16 width, UInt16 canvasWidth, Int16 xExtent) const
{
	fralloc +

	movem.l D0-D7/A0-A2, -(SP)		

	move.l foregroundBits, A0
	move.l backgroundBits, A1
	move.l drawBits, A2
	move.w yExtent, D0
	subq.w #1, D0
	move.w colQuads, D1
	subq.w #1, D1
	moveq.l #0, D6
	move.w xExtent, D6
	moveq.l #0, D4
	move.w width, D4
	sub.l D6, D4
	moveq.l #0, D5
	move.w canvasWidth, D5
	sub.l D6, D5

rowloop:
	move.w D1, D2

colloop:
	move.l (A2), D3
	move.l (A1)+, D7
	and.l   D7, D3
	or.l   (A0)+, D3
	move.l D3, (A2)+	

	move.l (A2), D3
	move.l (A1)+, D7
	and.l   D7, D3
	or.l   (A0)+, D3
	move.l D3, (A2)+	

dbra.w  D2, colloop

	add.l 	D4, A0
	add.l	D4, A1
	add.l   D5, A2
	
dbra.w  D0, rowloop

	movem.l (SP)+, D0-D7/A0-A2

	frfree

	rts	
}
#endif




Direct8bppSolidBitmapArray::Direct8bppSolidBitmapArray(const Canvas *canvas, UInt16 numBitmaps, UInt16 bitmapID)
	: BitmapArray(numBitmaps), bitmaps(new Direct8bppSolidBitmap[numBitmaps]),
	bitmapFtrFirstNum(bitmapID)
{
	Err error = errNone;

	// Determine bitmap extents
	MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID);
	ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
	BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));

	BmpGlueGetDimensions(resP, &width, &height, NULL);

	MemPtrUnlock(resP);
	DmReleaseResource( resH );

	// Prepare all bitmaps
	RectangleType bitmapRect;
	bitmapRect.topLeft.x = bitmapRect.topLeft.y = 0;
	bitmapRect.extent.x = width; bitmapRect.extent.y = height;

	const WinHandle buffer = WinCreateOffscreenWindow(width, height, screenFormat, &error);
	WinSetDrawWindow(buffer);

	WinSetForeColor(255);
	WinSetBackColor(0);
	
	for(UInt16 i = 0; i < numBitmaps; i++)
	{
		Direct8bppSolidBitmap *const bitmap = const_cast<Direct8bppSolidBitmap *>(&bitmaps[i]);
		// Paint sprite-bitmap to offscreen window
		MemHandle resH = DmGet1Resource( bitmapRsc, bitmapID + i);
		ErrNonFatalDisplayIf( !resH, "Missing bitmap" );
		BitmapType *const resP = static_cast<BitmapType *>(MemHandleLock(resH));
		WinDrawBitmap (resP, 0,0);
		error = BlitterSupport::copyBitmap(buffer, bitmapID + i, width, height, static_cast<BitmapType **>(NULL), reinterpret_cast<void **>(&(bitmap->bitmapBits)));
	
		MemPtrUnlock(resP);
		DmReleaseResource( resH );
	}

	WinSetDrawWindow(canvas->drawWindow);
	WinDeleteWindow(buffer, false);
}



Direct8bppSolidBitmapArray::~Direct8bppSolidBitmapArray()
{
	for (UInt16 i = 0; i < numBitmaps; i++)
	{
		FtrPtrFree(Device::creatorID, bitmapFtrFirstNum + i);
	}

	delete [] bitmaps;
}



void Direct8bppSolidBitmapArray::draw(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	const Direct8bppSolidBitmap *const bitmap = &bitmaps[index];

	// Determine memory address of bits
	UInt8 *foregroundBits = bitmap->bitmapBits;

	UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Clipping
	Int16 xOffset = 0;
	Int16 xExtent = width;
	if (x < 0)
	{
		xOffset = -x;
		xExtent -= xOffset;
	}
	
	if (x + width > canvas->width)
	{
		xExtent -= ((x + width) - canvas->width);
	}

	if (xExtent <= 0)
	{
		return;
	}
	

	Int16 yOffset = 0;
	Int16 yExtent = height;
	if (y < 0)
	{
		yOffset = 0 - y;
		yExtent -= yOffset;
	}
	
	if (y + height > canvas->height)
	{
		yExtent -= ((y + height) - canvas->height);
	}

	if (yExtent <= 0)
	{
		return;
	}
	

	// Paint
	UInt16 sourceOdd;
	{
		UInt16 sourceOffset = (width * yOffset + xOffset);
		foregroundBits += sourceOffset;
		
		sourceOdd = sourceOffset & 1;
	}
	
	drawBits += (canvas->width * (y + yOffset) + x + xOffset);

	const UInt16 destOdd = (x + xOffset) & 1;

	UInt16 colQuads = (xExtent - sourceOdd) >> 3;
	const UInt16 colRemainder = (xExtent - sourceOdd) & 7;

	if (!sourceOdd && !destOdd && !colRemainder)
	{
		// Everything is perfect for a blindingly fast blit :-)
#ifdef MODE_M68K
#ifdef __GNUC__
/*
	%0 foregroundBits
	%1 drawBits, 
	%2 (yExtent - 1)
	%3 (colQuads - 1),
	%4 (width - xExtent)
	%5 canvas->width - xExtent
*/

	const UInt32 canvasRemainder = canvas->width - xExtent;
	const UInt32 remainWidth = width - xExtent;
	yExtent--;
	colQuads--;

	asm("
solidrowloop:
	move.w %3, %%d2;

solidcolloop:
	move.l (%0)+, (%1)+;
	move.l (%0)+, (%1)+;
dbra.w  %%d2, solidcolloop;

	add.l 	%4, %0;
	add.l   %5, %1;
	
dbra.w  %2, solidrowloop;
	" : 
	  : "a" (foregroundBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (remainWidth), "r" (canvasRemainder)
	  : "%d2");
#elif __MWERKS__
    	const UInt32 canvasRemainder = canvas->width - xExtent;
    	const UInt32 remainWidth = width - xExtent;

		for (UInt16 row = yExtent; row > 0; row--)
		{
			for (UInt16 col = colQuads; col > 0; col--)
			{
				(*reinterpret_cast<UInt32 *>(drawBits)) = (*reinterpret_cast<UInt32 *>(foregroundBits));		
				foregroundBits += 4;
				drawBits += 4;

				(*reinterpret_cast<UInt32 *>(drawBits)) = (*reinterpret_cast<UInt32 *>(foregroundBits));		
				foregroundBits += 4;
				drawBits += 4;
			}
		
			foregroundBits += remainWidth;
			drawBits += canvasRemainder;
		}
#else
#error You are using an unsupported compiler
#endif
#else
    	const UInt32 canvasRemainder = canvas->width - xExtent;
    	const UInt32 remainWidth = width - xExtent;

		for (Int16 row = yExtent; row > 0; row--)
		{
			for (UInt16 col = colQuads; col > 0; col--)
			{
				(*reinterpret_cast<UInt32 *>(drawBits)) = (*reinterpret_cast<UInt32 *>(foregroundBits));		
				foregroundBits += 4;
				drawBits += 4;

				(*reinterpret_cast<UInt32 *>(drawBits)) = (*reinterpret_cast<UInt32 *>(foregroundBits));		
				foregroundBits += 4;
				drawBits += 4;
			}
		
			foregroundBits += remainWidth;
			drawBits += canvasRemainder;
		}
#endif
	}
	else
	{
		// Welcome to a world of lameness :-(
		const UInt32 canvasRemainder = canvas->width - xExtent;
		const UInt32 remainWidth = width - xExtent;

		for (Int16 row = 0; row < yExtent; row++)
		{
			for (Int16 col = 0; col < xExtent; col++)
			{
				(*drawBits) = (*foregroundBits);
				foregroundBits++;
				drawBits++;
			}
	
			foregroundBits += remainWidth;
			drawBits += canvasRemainder;
		}
	}
}





Direct8bppSaveBufferArray::Direct8bppSaveBufferArray(UInt16 numBuffers, Coord width, Coord height)
    : SaveBufferArray(numBuffers, width, height), buffers(new Direct8bppSaveBuffer[numBuffers])
{
	// Round width to next multiple of 4
	UInt16 bufferWidth = width;
	if ((bufferWidth & 3) != 0)
	{
		bufferWidth += 4 - (bufferWidth & 3);
	}

	// Add another 4 bytes
	bufferWidth += 4;
	
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		Direct8bppSaveBuffer *const buffer = &buffers[i];
		const UInt32 bufferSize = bufferWidth * height;
		buffer->bufferBits = static_cast<UInt8 *>(MemPtrNew(bufferSize));
		if (buffer->bufferBits == NULL)
		{
			Device::panic("Not enough memory for save buffer.");
		}		
	}
}


Direct8bppSaveBufferArray::~Direct8bppSaveBufferArray()
{
	for (UInt16 i = 0; i < numBuffers; i++)
	{
		const Direct8bppSaveBuffer *const buffer = &buffers[i];
		if (buffer->bufferBits != NULL)
		{
			MemPtrFree(buffer->bufferBits);
		}
	}
	
	delete [] buffers;
}


void Direct8bppSaveBufferArray::save(const Canvas *canvas, UInt16 index, Coord x, Coord y) const
{
	Direct8bppSaveBuffer *const buffer = &buffers[index];

	UInt8 *bufferBits = buffer->bufferBits;
	const UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Clipping
	Int16 xOffset = 0;
	Int16 xExtent = width;
	if (x < 0)
	{
		xOffset = -x;
		xExtent -= xOffset;
	}
	
	if (x + width > canvas->width)
	{
		xExtent -= ((x + width) - canvas->width);
	}

	if (xExtent <= 0)
	{
		buffer->colQuads = 0;
		return;
	}

	Int16 spanLeftX = x + xOffset;

	// Force even start address
	if ((spanLeftX & 1) != 0)
	{
		spanLeftX--;
	}

	// Make sure, entire width is covered
	if (spanLeftX + xExtent < x + width)
	{
		xExtent &= ~3;
		xExtent += 4;
	}

	// Make sure, memory addresses will be even
	if ((xExtent & 3) != 0)
	{
		xExtent &= ~3;
		xExtent += 4;
	}

	// Make sure covered span doesn't go off the right edge of Canvas
	if (spanLeftX + xExtent > canvas->width)
	{
		spanLeftX = canvas->width - xExtent;		
	}
	

	Int16 yOffset = 0;
	Int16 yExtent = height;
	if (y < 0)
	{
		yOffset = 0 - y;
		yExtent -= yOffset;
	}
	
	if (y + height > canvas->height)
	{
		yExtent -= ((y + height) - canvas->height);
	}

	if (yExtent <= 0)
	{
		buffer->yExtent = 0;
		return;
	}
	
	// Calculate low-level parameters
	const UInt32 drawOffset = canvas->width * (y + yOffset) + spanLeftX;
	UInt16 colQuads = xExtent >> 2;
	const UInt32 canvasRemainder = canvas->width - xExtent;

	// Save low-level parameters in buffer metadata
	buffer->canvasRemainder = canvasRemainder;
	buffer->drawOffset = drawOffset;
	buffer->colQuads = colQuads;
	buffer->yExtent = yExtent;

	// Save drawbits to buffer
	drawBits += drawOffset;

#ifdef MODE_M68K

#ifdef __GNUC__
	yExtent--;
	colQuads--;

	asm("
saverowloop:
	move.w %3, %%d2;

savecolloop:
	move.l (%1)+, (%0)+;
dbra.w  %%d2, savecolloop;

	add.l   %4, %1;
	
dbra.w  %2, saverowloop;
	" : 
	  : "a" (bufferBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (canvasRemainder)
	  : "%d2");
#else
	// Codewarrior
	saveM68KAsm(bufferBits, drawBits, yExtent, colQuads, canvasRemainder);
#endif
#else
	// Unknown processor type
	for (UInt16 row = 0; row < yExtent; row++)
	{
		for (UInt16 col = 0; col < colQuads; col++)
		{
			(*((UInt32 *)bufferBits)) = (*((UInt32 *)drawBits));		
			bufferBits += 4;
			drawBits += 4;
		}

		drawBits += canvasRemainder;
	}
#endif
}


void Direct8bppSaveBufferArray::restore(const Canvas *canvas, UInt16 index) const
{
	const Direct8bppSaveBuffer *const buffer = &buffers[index];

	const UInt8 *bufferBits = buffer->bufferBits;
	UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);

	// Restore low-level metadata from buffer
	const UInt32 drawOffset = buffer->drawOffset;
	UInt16 colQuads = buffer->colQuads;
	const UInt32 canvasRemainder = buffer->canvasRemainder;
	Int16 yExtent = buffer->yExtent;

	if ((colQuads == 0) || (yExtent == 0))
	{
		return;
	}

	// Restore buffer to drawing canvas
	drawBits += drawOffset;

#ifdef MODE_M68K

#ifdef __GNUC__
	yExtent--;
	colQuads--;

	asm("
restorerowloop:
	move.w %3, %%d2;

restorecolloop:
	move.l (%0)+, (%1)+;
dbra.w  %%d2, restorecolloop;

	add.l   %4, %1;
	
dbra.w  %2, restorerowloop;
	" : 
	  : "a" (bufferBits), "a" (drawBits), "d" (yExtent), "r" (colQuads), "r" (canvasRemainder)
	  : "%d2");
#else
	// Codewarrior
	restoreM68KAsm(bufferBits, drawBits, yExtent, colQuads, canvasRemainder);
#endif
#else
	// Unknown processor
	for (UInt16 row = 0; row < yExtent; row++)
	{
		for (UInt16 col = 0; col < colQuads; col++)
		{
			(*((UInt32 *)drawBits)) = (*((UInt32 *)bufferBits));		
			bufferBits += 4;
			drawBits += 4;
		}

		drawBits += canvasRemainder;
	}
#endif
}


#ifdef __MWERKS__
asm void Direct8bppSaveBufferArray::saveM68KAsm(
	void *bufferBits, const void *drawBits,
	Int16 yExtent, UInt16 colQuads, UInt32 canvasRemainder
	) const
{
	fralloc +

	movem.l D0-D3/A0-A1, -(SP)

	move.l bufferBits, A0
	move.l drawBits, A1

	move.w yExtent, D0
	subq.w #1, D0
	move.w colQuads, D1
	subq.w #1, D1
	move.l canvasRemainder, D3	

rowloop:
	move.w D1, D2

colloop:
	move.l (A1)+, (A0)+

dbra.w  D2, colloop

	add.l 	D3, A1
	
dbra.w  D0, rowloop

	movem.l (SP)+, D0-D3/A0-A1

	frfree

	rts	
}


asm void Direct8bppSaveBufferArray::restoreM68KAsm(
	const void *bufferBits, void *drawBits,
	Int16 yExtent, UInt16 colQuads, UInt32 canvasRemainder
	) const
{
	fralloc +

	movem.l D0-D3/A0-A1, -(SP)

	move.l bufferBits, A0
	move.l drawBits, A1

	move.w yExtent, D0
	subq.w #1, D0
	move.w colQuads, D1
	subq.w #1, D1
	move.l canvasRemainder, D3	

rowloop:
	move.w D1, D2

colloop:
	move.l (A0)+, (A1)+

dbra.w  D2, colloop

	add.l 	D3, A1
	
dbra.w  D0, rowloop

	movem.l (SP)+, D0-D3/A0-A1

	frfree

	rts	
}
#endif






Direct8bppBlitter::Direct8bppBlitter(Canvas *canvas) : Blitter(canvas)
{
}


Direct8bppBlitter::~Direct8bppBlitter()
{
}



const BitmapArray *Direct8bppBlitter::prepareTransparentBitmaps(UInt16 numBitmaps, DmResID bitmapID, DmResID maskID) const
{
	return new Direct8bppTransparentBitmapArray(canvas, numBitmaps, bitmapID, maskID);
}


const BitmapArray *Direct8bppBlitter::prepareSolidBitmaps(UInt16 numBitmaps, DmResID bitmapID) const
{
	return new Direct8bppSolidBitmapArray(canvas, numBitmaps, bitmapID);
}








void Direct8bppBlitter::drawText(Color textColor, Color bgColor, FontID font, const Char *chars, Int16 len, Coord x, Coord y) const
{
	canvas->setForegroundColor(textColor);
	canvas->setBackgroundColor(bgColor);

	FntSetFont(font);	
	WinDrawChars(chars, len, x, y);
}	



void Direct8bppBlitter::fillRectangle(Color color, Coord topX, Coord topY, Coord bottomX, Coord bottomY) const
{
	// Clip
	if ((bottomY < 0) || (bottomX < 0) || (topX >= canvas->width) || (topY >= canvas->height))
	{
		return;
	}
		
	if (topX < 0)
		topX = 0;
	if (topY < 0)
		topY = 0;
	if (bottomX >= canvas->width)
		bottomX = canvas->width - 1;
	if (bottomY >= canvas->height)
		bottomY = canvas->height - 1;
	
	if ((topX & 1) || (((bottomX - topX) + 1) & 7))
	{
		// Criteria for high performance are not fulfilled...
		canvas->setForegroundColor(color);

		RectangleType rectangle;
		rectangle.topLeft.x = topX;
		rectangle.topLeft.y = topY;
		rectangle.extent.x = (bottomX - topX) + 1;
		rectangle.extent.y = (bottomY - topY) + 1;
		
		WinDrawRectangle(&rectangle, 0);
	}
	else if ((topX == 0) && (bottomX == canvas->width - 1))
	{
		// Rectangle covers entire screen line
		UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);
		drawBits += (topY * canvas->width);
		const UInt16 packets = (canvas->rowBytes / 160) * ((bottomY - topY) + 1);
		const UInt32 colorIdx = static_cast<UInt32>(color.getIndexedColor());
		const UInt32 colorUInt32 = (colorIdx << 24) + (colorIdx << 16) + (colorIdx << 8) + colorIdx;
		
#ifdef __GNUC__
    asm("movem.l %d0-%d7/%a0-%a6, -(%sp)");

    asm("move.l  %0, %%a1" : : "g" (drawBits));
    asm("move.w  %0, %%d0" : : "g" (packets));
    asm("move.l  %0, %%d1" : : "g" (colorUInt32));
	asm("subq.w #1, %D0");

    asm("
		move.l %d1, %d2
		move.l %d1, %d3
		move.l %d1, %d4
		move.l %d1, %d5
		move.l %d1, %d6
		move.l %d1, %d7
		move.l %d1, %a2
		move.l %d1, %a3
		move.l %d1, %a4

copyLoop:
         movem.l %d1-%d7/%a2-%a4, (%a1)
         movem.l %d1-%d7/%a2-%a4, 40(%a1)
         movem.l %d1-%d7/%a2-%a4, 80(%a1)
         movem.l %d1-%d7/%a2-%a4, 120(%a1)

         adda.l  #160, %a1

         dbra    %d0, copyLoop
    ");

    asm("movem.l (%sp)+, %d0-%d7/%a0-%a6");
#else
		for (UInt16 packetCount = packets; packetCount > 0; packetCount--)
		{
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;

			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
			(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
			drawBits += 4;
		}
#endif
	}
	else
	{
		UInt8 *drawBits = static_cast<UInt8 *>(canvas->drawBits);
		drawBits += (topY * canvas->width) + topX;
		const UInt16 xExtent = (bottomX - topX) + 1;
		const UInt16 packets = xExtent >> 3;
		const UInt16 yExtent = (bottomY - topY) + 1;
		const UInt32 colorIdx = static_cast<UInt32>(color.getIndexedColor());
		const UInt32 colorUInt32 = (colorIdx << 24) + (colorIdx << 16) + (colorIdx << 8) + colorIdx;

		const UInt16 drawBitsIncrement = canvas->width - xExtent;
		for (UInt16 rows = yExtent; rows > 0; rows--)
		{
			for(UInt16 cols = packets; cols > 0; cols--)
			{
				(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
				drawBits += 4;
				(*(reinterpret_cast<UInt32 *>(drawBits))) = colorUInt32;
				drawBits += 4;
			}
			
			drawBits += drawBitsIncrement;
		}
	}
}




SaveBufferArray *Direct8bppBlitter::prepareSaveBuffers(UInt16 numBuffers, Coord width, Coord height) const
{
    return new Direct8bppSaveBufferArray(numBuffers, width, height);    
}
