/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: DirectDoubleBuffer.cpp,v 1.4 2003/05/29 15:45:25 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

/*
 * TODO: This will not work on hires devices in compatibility mode, b/c of mismatch between Canvas::width and 
 *       physical display size. Adapt to rowbytes.
 */


#include "./HostControl.h"
#include <PalmOS.h>

#include "Customization.h"
#include <BmpGlue.h>
#include "gfx/DirectDoubleBuffer.h"
#include "Utilities.h"
#include "Device.h"
#include "gfx/BlitterSupport.h"
#include "gfx/Canvas.h"


using namespace prvrazor;


Boolean DirectDoubleBuffer::isSupported(const Canvas *targetCanvas)
{
	// This is currently done for safety reasons. It is also unlikely that optimized m68k code
	// would bring speed advantages on ARM devices.
	if (Device::supports50)
	{
		return false;
	}

	const Coord width = targetCanvas->width;
	const Coord height = targetCanvas->height;

	// DirectDoubleBuffer is supported on M68K (emulating) devices with a 160x160 or 320x320 logical & physical display,
	// and supported memory layout.
	#ifdef MODE_M68K
		HostTraceOutputTL2(appErrorClass, "width %hd   height %hd", width, height);

		if (width != height)
		{
			return false;   // Display not square
		}
		
		if ((width != 160) && (width != 320))
		{
			return false;   // Display not 160x160 or 320x320
		}


		if (targetCanvas->rowBytes == ((width * Display::depth) / 8))
		{
			HostTraceOutputTL(appErrorClass, "VRAM layout supported");
			return true;
		}
		else
		{
			HostTraceOutputTL(appErrorClass, "VRAM layout unsupported");
			return false;
		}		
	#else
		// DirectDoubleBuffer cannot provide native code for any other CPU.
		return false;
	#endif
}


DirectDoubleBuffer::DirectDoubleBuffer(const Canvas *sourceCanvas, const Canvas *targetCanvas, const ViewportProperties& viewport)
{
	this->displayBits   = targetCanvas->drawBits;
	this->viewportBits = sourceCanvas->drawBits;

    targetWidth = targetCanvas->width;

	screenLineSize   = targetCanvas->rowBytes;

	this->yOffsetBytes = viewport.region.topLeft.y * screenLineSize;
	viewportExtentY = viewport.region.extent.y;

	// Determine the parameters of the "copy unit".
	if (screenLineSize >= PACKETSIZE)
	{
		linesPerCopyUnitShift = 0;
		packetsPerCopyUnit = screenLineSize / PACKETSIZE;
	}
	else
	{
		UInt16 linesPerCopyUnit   = PACKETSIZE / screenLineSize;
		packetsPerCopyUnit = 1;

		switch(linesPerCopyUnit)
		{
			case 2:
				linesPerCopyUnitShift = 1;
				linesPerCopyUnitMask  = 1;
				break;
			case 4:
				linesPerCopyUnitShift = 2;
				linesPerCopyUnitMask  = 3;
				break;
			case 8:
				linesPerCopyUnitShift = 3;
				linesPerCopyUnitMask  = 7;
				break;
		}

		ErrNonFatalDisplayIf((viewportExtentY & linesPerCopyUnitMask) != 0, "Illegal viewport height");
	}


	dontCopy = true;
}

	
	
DirectDoubleBuffer::~DirectDoubleBuffer()
{
}


	
void DirectDoubleBuffer::prepareCopy(RectangleType &copyBounds)
{	
	// Don't show stuff off screen
	// TODO: Are simpler comparisons possible? How much is already done by Display::endDrawViewport?
	if 
	(
		(copyBounds.extent.y == 0) || (copyBounds.topLeft.y > viewportExtentY) || 
		(copyBounds.extent.x == 0) || (copyBounds.topLeft.x > targetWidth) ||
		((copyBounds.topLeft.y + copyBounds.extent.y) <= 0) || ((copyBounds.topLeft.x + copyBounds.extent.x) <= 0)
	)
	{
		dontCopy = true;  // There is nothing on the screen.
	}
	else
	{
		dontCopy = false;
		
		copyTop = max(0, copyBounds.topLeft.y);
		copyExtent = min(viewportExtentY, copyBounds.topLeft.y + copyBounds.extent.y) - copyTop;
	}
}


void DirectDoubleBuffer::copy()
{
	if (!dontCopy)
	{
		this->copyScreen();
	}
}




#ifdef MODE_M68K

#ifdef __GNUC__
void DirectDoubleBuffer::copyScreen()
{
	UInt16 lineCount = (UInt16)(copyExtent);
	UInt16 packetCount;

	if (linesPerCopyUnitShift == 0)
	{
		packetCount = lineCount * packetsPerCopyUnit;
	}
	else
	{
		// Make sure, that the number of copied lines is a multiple of
		// the lines per copy unit.
		if ((lineCount & linesPerCopyUnitMask) != 0)
		{
			lineCount = ((lineCount >> linesPerCopyUnitShift) + 1) << linesPerCopyUnitShift;
		}
		
		packetCount = lineCount >> linesPerCopyUnitShift;
	}

	Coord topLine = copyTop;
	if ((topLine + lineCount) > viewportExtentY)
	{
		topLine = viewportExtentY - lineCount;
	}
	
	UInt32 offset = topLine * screenLineSize;

	void *fromPtr = (UInt8 *)viewportBits + offset;
	void *toPtr   = (UInt8 *)displayBits + offset + yOffsetBytes;
    asm("movem.l %d0-%d7/%a0-%a6, -(%sp)");

    asm("move.l  %0, %%a0" : : "g" (fromPtr));
    asm("move.l  %0, %%a1" : : "g" (toPtr));
    asm("move.w  %0, %%d0" : : "g" (packetCount));
	asm("subq.w #1, %D0");

    asm("
copyLoop:
         movem.l (%a0)+, %d1-%d7/%a2-%a4
         movem.l %d1-%d7/%a2-%a4, (%a1)

         movem.l (%a0)+, %d1-%d7/%a2-%a4
         movem.l %d1-%d7/%a2-%a4, 40(%a1)

         movem.l (%a0)+, %d1-%d7/%a2-%a4
         movem.l %d1-%d7/%a2-%a4, 80(%a1)

         movem.l (%a0)+, %d1-%d7/%a2-%a4
         movem.l %d1-%d7/%a2-%a4, 120(%a1)

         adda.l  #160, %a1                    

         dbra    %d0, copyLoop
    ");

    asm("movem.l (%sp)+, %d0-%d7/%a0-%a6");
}


#else   // Assuming CodeWarrior


void DirectDoubleBuffer::copyScreen()
{
	UInt16 lineCount = (UInt16)(copyExtent);
	UInt16 packetCount;

	if (linesPerCopyUnitShift == 0)
	{
		packetCount = lineCount * packetsPerCopyUnit;
	}
	else
	{
		// Make sure, that the number of copied lines is a multiple of
		// the lines per copy unit.
		if ((lineCount & linesPerCopyUnitMask) != 0)
		{
			lineCount = ((lineCount >> linesPerCopyUnitShift) + 1) << linesPerCopyUnitShift;
		}
		
		packetCount = lineCount >> linesPerCopyUnitShift;
	}

	Coord topLine = copyTop;
	if ((topLine + lineCount) > viewportExtentY)
	{
		topLine = viewportExtentY - lineCount;
	}
	
	UInt32 offset = topLine * screenLineSize;

	void *fromPtr = (UInt8 *)viewportBits + offset;
	void *toPtr   = (UInt8 *)displayBits + offset + yOffsetBytes;

	copyScreenLinesM68KAsm(fromPtr, toPtr, packetCount);
}


asm void DirectDoubleBuffer::copyScreenLinesM68KAsm(const void *fromPtr, void *toPtr, UInt16 packetCount)
{
	fralloc +

	movem.l D0-D7/A0-A6, -(SP)		

	move.l fromPtr, A0
	move.l toPtr,   A1
	move.w packetCount, D0
	subq.w #1, D0

loop:
	movem.l (A0)+, D1-D7/A2-A4
	movem.l D1-D7/A2-A4, (A1)
	movem.l (A0)+, D1-D7/A2-A4
	movem.l D1-D7/A2-A4, 40(A1)
	movem.l (A0)+, D1-D7/A2-A4
	movem.l D1-D7/A2-A4, 80(A1)
	movem.l (A0)+, D1-D7/A2-A4
	movem.l D1-D7/A2-A4, 120(A1)
	add.l   #160, A1
	
	dbra.w  D0, loop

	movem.l (SP)+, D0-D7/A0-A6

	frfree

	rts
}
#endif  // __GNUC__

#else // other CPU

void DirectDoubleBuffer::copyScreen()
{
	ErrNonFatalDisplay("DirectDoubleBuffer::copyScreen: Unsupported CPU");
}

#endif // CPU selection
