/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: PalmOSCanvas.cpp,v 1.5 2003/05/30 12:30:20 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/

/*
 * TODO:
 * - Work around faulty (too smart) HandEra offscreen window creation (create bitmap, then Window)
 *   Maybe this should be done on all devices.
 * - Should GenericFormat be used? That way the direct blitters could be used in a fully sanctioned way for extra speed.
 * - Can WinSetColors be used for legacy devices?
 */


#include "Customization.h"

#include "./HostControl.h"
#include <PalmOS.h>
#include <BmpGlue.h>
#include "Device.h"
#include "gfx/Display.h"
#include "gfx/PalmOSCanvas.h"
#include "gfx/BlitterSupport.h"
#include "gfx/PalmOSBlitter.h"
#ifndef MODE_PURITY
#include "gfx/Direct4bppBlitter.h"
#include "gfx/Direct8bppBlitter.h"
#endif

#ifndef USE_SDK50
#define nativeFormat screenFormat
#endif

using namespace prvrazor;


PalmOSCanvas::PalmOSCanvas(Coord width, Coord height, Canvas::CoordSystem coordSystem) 
    : Canvas(width, height, coordSystem)
{	
	UInt16 error;
	// Parameters to WinCreateOffscreenWindow are in LOGICAL coordinates!!!
	offscreenWindow = WinCreateOffscreenWindow(width, height, nativeFormat, &error);
	if (error != errNone)
	{
		Device::panic("Cannot create double buffer.");
	}

	drawWindow = offscreenWindow;
	_drawBitmap = BlitterSupport::getWindowBitmap(drawWindow);
	drawBits = BlitterSupport::getWindowBits(drawWindow);

	// TODO: Is this distinction neccessary?
	if (Device::supports35)
	{
		BmpGlueGetDimensions(_drawBitmap, NULL, NULL, &rowBytes);
	}
	else
	{
		rowBytes = (cols * Display::depth) / 8;
	}


#ifndef MODE_PURITY
	if (Device::supports50)
	{
		blitter = new PalmOSBlitter(this);
	}
	else if (Display::depth == 4)
	{
		blitter = new Direct4bppBlitter(this);
	}
	else if (Display::depth == 8)
	{
		blitter = new Direct8bppBlitter(this);
	}
	else
#endif
	{
		blitter = new PalmOSBlitter(this);
	}
}


PalmOSCanvas::PalmOSCanvas() 
    : Canvas(Display::cols, Display::rows, Canvas::COORDS_NATIVE)
{
	offscreenWindow = NULL;

	drawWindow = WinGetDisplayWindow();
	_drawBitmap = BlitterSupport::getWindowBitmap(drawWindow);
	drawBits = BlitterSupport::getWindowBits(drawWindow);

	// TODO: Is this distinction neccessary?
	if (Device::supports35)
	{
		BmpGlueGetDimensions(_drawBitmap, NULL, NULL, &rowBytes);
	}
	else
	{
		rowBytes = (cols * Display::depth) / 8;
	}

#ifndef MODE_PURITY
	if (Device::supports50)
	{
		blitter = new PalmOSBlitter(this);
	}
	else if (Display::depth == 4)
	{
		blitter = new Direct4bppBlitter(this);
	}
	else if (Display::depth == 8)
	{
		blitter = new Direct8bppBlitter(this);
	}
	else
#endif
	{
		blitter = new PalmOSBlitter(this);
	}
}
	
	
PalmOSCanvas::~PalmOSCanvas()
{
	if (blitter != NULL)
	{
		delete blitter;
	}
	
	if (offscreenWindow != NULL)
	{
		WinDeleteWindow(offscreenWindow, false);
	}
}


void PalmOSCanvas::setOSCanvas() const
{
    WinSetDrawWindow(drawWindow);
}

Boolean PalmOSCanvas::setForegroundColor(const Color color) const
{
    return PalmOSCanvas::_setForegroundColor(color);
}

Boolean PalmOSCanvas::setBackgroundColor(const Color color) const
{
    return PalmOSCanvas::_setBackgroundColor(color);
}



Boolean PalmOSCanvas::_setForegroundColor(const Color color)
{
	Boolean result = true;

	const IndexedColorType index = color.getIndexedColor();

	switch (Display::depth)
	{
		case 1:
		case 2:
		case 4:
			if (Device::supports35)
			{
				WinSetForeColor(index);
				WinSetTextColor(index);
			}
			else
			{
				// Pre-OS 3.5 devices only support B/W with their API. Returning true means "black", false means "white"
				// Indices are sorted from 0 = white ... 2^depth = black.
				result = (index > (((2 << Display::depth) >> 1) - 1));
			}
			break;
		case 8:
			WinSetForeColor(index);
			WinSetTextColor(index);
			break;
		case 16:
			break;
	}

	return result;
}


Boolean PalmOSCanvas::_setBackgroundColor(const Color color)
{
	Boolean result = true;

	const IndexedColorType index = color.getIndexedColor();

	switch (Display::depth)
	{
		case 1:
		case 2:
		case 4:
			if (Device::supports35)
			{
				WinSetBackColor(index);
			}
			else
			{
				// Pre-OS 3.5 devices only support B/W with their API. Returning true means "black", false means "white"
				// Indices are sorted from 0 = white ... 2^depth = black.
				result = (index > (((2 << Display::depth) >> 1) - 1));
			}
			break;
		case 8:
			WinSetBackColor(index);
			break;
		case 16:
			break;
	}

	return result;
}
