/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: PalmOSDisplay.cpp,v 1.5 2003/05/30 12:30:21 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/


/*
 * TODO:
 * - Set nativeCoordinates before each draw op and set back to normal afterwards, instead of trying to detect when the UI
 *   layer wants to draw sth. (see KB 341: Q: Even though I have set the coordinate system to native why does the system keeps setting it back to single density?)
 *
 * SEE:
 * - KB article 341
 */

#include "Customization.h"

#include "./HostControl.h"
#include <PalmOS.h>
#include <PalmOsGlue.h>
#include "gfx/PalmOSDisplay.h"
#include "gfx/PalmOSCanvas.h"
#include "Device.h"

#ifndef USE_SDK50
#define nativeFormat screenFormat
#endif

using namespace prvrazor;


PalmOSDisplay::PalmOSDisplay() : Display()
{	
#ifndef MODE_PURITY
	directDoubleBuffer = NULL;
#endif

	displayCaptured = false;
	saveDrawState(originalDrawState);

    initPhysicalDisplay();
	displayCanvas = new PalmOSCanvas();
}
	
	
PalmOSDisplay::~PalmOSDisplay()
{
#ifndef MODE_PURITY
	if (directDoubleBuffer != NULL)
	{
		delete directDoubleBuffer;
	}
#endif
}

	
void PalmOSDisplay::endDrawViewport(RectangleType *bounds)
{
	Display::endDrawViewport(bounds);

#ifndef MODE_PURITY
	if (directDoubleBuffer != NULL)
	{
		directDoubleBuffer->prepareCopy(copyBounds);
	}
	else
#endif
	{
		// Align the x coordinate with the nearest word to speed up the WinCopyRectangle.
		UInt16 offset = copyBounds.topLeft.x & 0x000F;
		if (offset)
		{
			copyBounds.topLeft.x -= offset;
			copyBounds.extent.x += offset;
		}
		
		// Push the extent out to fill the entire word.
		copyBounds.extent.x = (copyBounds.extent.x + 0x0F) & 0xFFF0; 
	}
}


void PalmOSDisplay::showViewport()
{
#ifndef MODE_PURITY
	if (directDoubleBuffer != NULL)
	{
		directDoubleBuffer->copy();
	}
	else
#endif
	{
		if (Device::supports35 && (Display::depth == 1))
		{
			WinSetForeColor(1);
			WinSetBackColor(0);		
		}

		// Copy the screenBuffer
		WinCopyRectangle (viewportCanvas->drawWindow, 0, &copyBounds, 
			copyBounds.topLeft.x, 
			copyBounds.topLeft.y + viewport.region.topLeft.y, 
			winPaint);  
	}

#ifdef MODE_DEBUG_BOUNDS
	// The current bounds should match the bounds which are returned by Screen::drawAction()
	RectangleType adjustedCurrentBounds;
	RctCopyRectangle(&currentBounds, &adjustedCurrentBounds);
	RctOffsetRectangle(&adjustedCurrentBounds, 0, viewport.region.topLeft.y);
	WinSetForeColor(1);
	WinDrawRectangleFrame(simpleFrame, &adjustedCurrentBounds);

	// Last bounds are the current bounds from the previous screen
	RectangleType adjustedLastBounds;
	RctCopyRectangle(&lastBounds, &adjustedLastBounds);
	RctOffsetRectangle(&adjustedLastBounds, 0, viewport.region.topLeft.y);
	WinSetForeColor(2);
	WinDrawRectangleFrame(simpleFrame, &adjustedLastBounds);

	// Copy bounds is a merged version of last bounds and current bounds.
	RectangleType adjustedCopyBounds;
	RctCopyRectangle(&copyBounds, &adjustedCopyBounds);
	RctOffsetRectangle(&adjustedCopyBounds, 0, viewport.region.topLeft.y);
	WinSetForeColor(3);
	WinDrawRectangleFrame(simpleFrame, &adjustedCopyBounds);
#endif
}



void PalmOSDisplay::initDisplay(const DisplayProperties& displayProperties)
{
	// Capture presentation draw state
	saveDrawState(presentationDrawState);	

    // Set new draw state and logical size of display
    if (Device::supports50)
    {
        switch(displayProperties.coordSystem)
        {
            case Canvas::COORDS_NATIVE:
                WinSetCoordinateSystem(kCoordinatesNative);
            break;
            case Canvas::COORDS_STANDARD:
                WinSetCoordinateSystem(kCoordinatesStandard);
            break;
        }
    }

    displayCanvas->setCoordSystem(displayProperties.coordSystem);
}


void PalmOSDisplay::captureDisplay()
{
	if (!displayCaptured)
	{
		restoreDrawState(presentationDrawState);
		displayCaptured = true;
	}
}


void PalmOSDisplay::releaseDisplay()
{
	if (displayCaptured)
	{
		restoreDrawState(originalDrawState);
		displayCaptured = false;
	}
}


void PalmOSDisplay::saveDrawState(DrawStateType& drawState)
{
	if (Device::supports50)
	{
#ifdef USE_SDK50
		drawState.coordinateSystem = WinGetCoordinateSystem();
#else // Device is OS5, but SDK is not at least SDK 5
		Device::panic("This version does not support OS5 devices.");
#endif  // USE_SDK50
	}
}


void PalmOSDisplay::restoreDrawState(const DrawStateType& drawState)
{
	if (Device::supports50)
	{
#ifdef USE_SDK50
		WinSetCoordinateSystem(drawState.coordinateSystem);
#else // Device is OS5, but SDK is not at least SDK 5
		Device::panic("This version does not support OS5 devices.");
#endif  // USE_SDK50
	}
}



void PalmOSDisplay::resetDisplay()
{
   	WinScreenMode(winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
}


void PalmOSDisplay::createViewportCanvas(const ViewportProperties& viewport)
{
	viewportCanvas = new PalmOSCanvas(viewport.region.extent.x, viewport.region.extent.y, 
		displayCanvas->coordSystem);

#ifndef MODE_PURITY
	// We got both canvases now, so let's see if the DirectDoubleBuffer can be used.
	if (directDoubleBuffer != NULL)
	{
		delete directDoubleBuffer;
	}
	
	if (DirectDoubleBuffer::isSupported(displayCanvas))
	{
		directDoubleBuffer = new DirectDoubleBuffer(viewportCanvas, displayCanvas, viewport);
	}
#endif
}



void PalmOSDisplay::initPhysicalDisplay()
{
    MemHandle resH = DmGetResource( (DmResType)'Tbmt', 1000);
    ErrNonFatalDisplayIf(!resH, "Bitmap metadata not found!");
    void *resP = MemHandleLock(resH);

    UInt8 *metaBytes = static_cast<UInt8 *>(resP);

    // Determine supported depths
	UInt32 supportedDepths;
	WinScreenMode(winScreenModeGetSupportedDepths, NULL, NULL, &supportedDepths, NULL);		

    // Determine current display density
    Display::Density displayDensity;
	if (Device::supports50)
	{
#ifdef USE_SDK50
		// Palm OS 5.0
		UInt32 physicalWidth;
		WinScreenGetAttribute(winScreenWidth, &physicalWidth);
		UInt32 physicalHeight;
		WinScreenGetAttribute(winScreenHeight, &physicalHeight);
        Display::rows = physicalHeight;
        Display::cols = physicalWidth;

        UInt32 currentDensity;
        WinScreenGetAttribute(winScreenDensity, &currentDensity);
        switch(currentDensity)
        {
            case kDensityLow:
                displayDensity = Display::DENSITY_SINGLE;
                break;
            case kDensityDouble:
                displayDensity = Display::DENSITY_DOUBLE;
                break;
            default:
                Device::panic("Your device has a display with unsupported properties.");
        }
#else
    	Device::panic("This version does not support OS5 devices.");
#endif
    }
    else
    {
        Display::rows = 160;
        Display::cols = 160;
        displayDensity = Display::DENSITY_SINGLE;
    }

    Display::density = displayDensity;

    ///////////////////////////////////////////////////////////////////////////////////
    // Iterate over the resource metadata sets, trying to find a match
    ///////////////////////////////////////////////////////////////////////////////////
    bool found = false;    
    while ((*metaBytes != 0x00) && !found)
    {
        const UInt8 desiredDensity = *metaBytes;
        const UInt8 desiredDepth = *(metaBytes + 1);

        bool densityOk = false;
        if ((displayDensity == Display::DENSITY_SINGLE) && (desiredDensity == 1))
        {
            densityOk = true;
        }
        else if (displayDensity == Display::DENSITY_DOUBLE)
        {
            densityOk = true;
        }

        bool depthOk = (supportedDepths & (1 << (desiredDepth - 1)));

        if (densityOk && depthOk)
        {
        	// Set selected display depth	
            Display::depth = desiredDepth;

        	UInt32 currentDisplayDepth;
        	WinScreenMode(winScreenModeGet, NULL, NULL, &currentDisplayDepth, NULL);
        	if (currentDisplayDepth != Display::depth)
        	{
        		const Err error = WinScreenMode(winScreenModeSet, NULL, NULL, &Display::depth, NULL);
                if (error != errNone)
                {
             		Device::panic("Unable to set display depth");
             	}
         	}


        	// Is direct color supported?
        	// See http://oasis.palm.com/dev/kb/faq/2449.cfm
        	Display::directColor = false;
        	if ((Display::depth >= 16) && Device::supports35 && Device::supportsTrap(sysTrapWinSetForeColorRGB))
        	{
        		Display::directColor = true;
        	}

            Display::colorMode = (Display::depth >= 8);

            found = true;
        }                

        metaBytes += 2;
    }
    
    
    MemPtrUnlock(resP);
    DmReleaseResource(resH);
    
    if (!found)
    {
        Device::panic("This version does not support your device's display properties.");
    }
}
