/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Device.h,v 1.4 2003/05/30 12:18:21 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/


#ifndef DEVICE_H
#define DEVICE_H

#include <PalmOS.h>
#include "RazorSections.h"

/**
 * Low level device operations.
 */
class Device
{
	public:

	/**
	 * Initialize the Device subsystem. This operation needs to be invoked
	 * before all others.
	 */
	static void init() SEC_RAZOR_INIT;

	/**
	 * Release the Device subsystem. This operation needs to be invoked
	 * before application shutdown.
	 */
	static void release() SEC_RAZOR_INIT;
	
	/**
	 * Display an error message to the user, then exit the application.
	 * Invoke this operation in case of an unresolvable runtime problem
	 * that is not your fault (e.g. low memory). Use the macros from the
	 * ErrorManager for all problems which are your fault (i.e. which
	 * will be ironed out in the release version).
	 */
	static void panic(const Char *message) SEC_RAZOR_INIT;

	/**
	 * Check that the ROM version is meeting your minimum requirements.
	 *
	 * @param requiredVersion minimum ROM version required (see sysFtrNumROMVersion in SystemMgr.h for format)
	 * @param launchFlags flags that indicate if the application UI is initialized.
	 *
	 * @return sysErrRomIncompatible, or errNone.
	 */
	static Err romVersionCompatible(UInt32 minVersion, UInt32 maxVersion, UInt16 launchFlags);

	/// Does the Palm device support Palm OS 3.0 Feature Set?
	static Boolean supports30;
	/// Does the Palm device support Palm OS 3.1 Feature Set?
	static Boolean supports31;
	/// Does the Palm device support Palm OS 3.5 Feature Set?
	static Boolean supports35;
	/// Does the Palm device support Palm OS 4.0 Feature Set?
	static Boolean supports40;
	/// Does the Palm device support Palm OS 5.0 Feature Set?
	static Boolean supports50;
	
	/// Is the device equipped with an ARM processor?
	static Boolean isArmed;

	/**
	 * Is the specified trap supported by the device?
	 */
	static Boolean supportsTrap(UInt16 trapNum);	


	/// The Creator ID of the currently running app
	static UInt32 creatorID;
	/// The application version of the currently running app
	static UInt16 appVersionNum;


	private:

	static UInt32 romVersion;	
	// Stage is specified as sysROMStageDevelopment, because that is returned by some POSE ROMs.
	static const UInt32 RomVersion30 = sysMakeROMVersion(3,0,0,sysROMStageDevelopment,0);
	static const UInt32 RomVersion31 = sysMakeROMVersion(3,1,0,sysROMStageDevelopment,0);
	static const UInt32 RomVersion35 = sysMakeROMVersion(3,5,0,sysROMStageDevelopment,0);
	static const UInt32 RomVersion40 = sysMakeROMVersion(4,0,0,sysROMStageDevelopment,0);
	static const UInt32 RomVersion50 = sysMakeROMVersion(5,0,0,sysROMStageDevelopment,0);
};


#endif
