/*********************************************************************************
 *
 * Razor! Engine - A modular C++ presentation engine
 *
 * $Id: Screen.h,v 1.5 2003/05/31 20:05:40 teacy Exp $
 *
 * Copyright (c) 2000-2003 Tilo Christ. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the  "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software 
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all 
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 **********************************************************************************/


#ifndef SCREEN_H
#define SCREEN_H

#include <PalmOS.h>
#include "StateDescriptor.h"
#include "gfx/Display.h"


/**
 * Screen defines the application specific behavior for a single screen.
 * <br>
 * Examples could be: A highscore screen, a game play screen, a bonus level screen,
 * a weapon shop screen, etc.
 * <br>
 * The order in which the Screens are to appear is defined by the ActionEngine.
 * <br>
 * The most important operations are the nextPeriod() operation which
 * is invoked in regular intervals and is supposed to update the
 * internal state of the Screen. And the drawAction() operation which is
 * invoked right afterwards, and is responsible for drawing a
 * visual representation of the current state onto the viewport Canvas.
 */
class Screen
{
	public:
/// @name Lifecycle management
//@{
	/**
	 * Construct a new Screen.
	 * <br><em>Caution: This is invoked BEFORE any Canvas or Display is properly set up!</em>
	 * This makes it impossible to initialize graphics assets in the constructor. Use init() instead.
	 */
	Screen() SEC_GAME;

	/**
	 * Destroy the Screen
	 */
	virtual ~Screen() SEC_GAME = 0;


	/**
	 * Initialize the internal state of the Screen, and prepare all assets (bitmaps, etc.)
	 * <br>
	 * <em>CAUTION!!! All bitmaps need to be prepared for the Canvas on which they will
	 * later be painted. No mixing is allowed!</em>
	 */
	virtual void init(const Canvas *displayCanvas, const Canvas *viewportCanvas) SEC_GAME = 0;
//@}

	
/// @name Control flow
//@{
	/**
	 * Progress to the next period. Update the internal state of your
	 * Screen, but <em>DO NOT DRAW ANYTHING!</em>. Only draw
	 * in your drawXXX() operations.
	 *
	 * @return True - if the Screen wants to be exited
	 *         False - if the Screen wants to remain the active Screen
	 */
	virtual Boolean nextPeriod() SEC_GAME = 0;

	/**
	 * In what intervals shall this screen be updated? 
	 *
	 * @return the length of the interval in 1/100 seconds (default is 4: 100/4 = 25fps).
	 */
	virtual UInt32 getAdvanceTimeInterval() SEC_GAME;
	
	/**
	 * How long shall the engine freeze the screen after resuming from a pause?
	 *
	 * @return the duration in 1/100 seconds (default is 0).
	 */
	virtual UInt32 getPauseLengthBeforeResuming() SEC_GAME;
//@}

	
/// @name Visual representation
//@{
    virtual void getDisplayProperties(DisplayProperties& displayProperties) const SEC_GAME;


	/**
	 * Return the viewport properties that the screen needs to operate.
	 * <br>
	 * This is invoked after the Display has been fully initialized, so it is OK to
	 * query Display properties within this operation.
	 * <br>
	 * <b>About the viewport rectangle</b>
	 * <br>
	 * Currently, only vertical limits are taken into account. Also, due
	 * to technical limitations, the height of the viewport needs to be a multiple of 8.
	 * <em>Ignoring this rule will lead to crashes on greyscale devices!</em>	 
	 * 
	 * @param displayCanvas the display canvas, which can be queried for information to calculate viewport properties
	 * @param viewport the viewport properties, which need to be filled in.
	 */
	virtual void getViewportProperties(const Canvas *displayCanvas, ViewportProperties& viewport) const SEC_GAME = 0;


	/**
	 * Draw the background of the viewport canvas. 
	 * It is assumed that this operation fills the entire viewport region with an image.
	 */
	virtual void drawBackground(const Canvas *viewportCanvas) const SEC_GAME = 0;


	/**
	 * Draw the visual appearance of the action onto the viewport canvas.
     *
	 * @param dirtyRectangle this needs to be filled in with the coordinates of the area
	 *               that was actually modified by the paint operations.<br><b>Failing to do so will
	 *               result in incomplete display updates</b>
	 */
	virtual void drawAction(const Canvas *viewportCanvas, RectangleType *dirtyRectangle) SEC_GAME = 0;	


	/**
	 * Draw any static imagery which lies outside the viewport region.
	 */
	virtual void drawFixedOverlays(const Canvas *displayCanvas) const SEC_GAME = 0;
	

	/**
	 * Draw any dynamic imagery which lies outside the viewport region.
	 */
	virtual void drawDynamicOverlays(const Canvas *displayCanvas) const SEC_GAME = 0;
//@}


/// @name Input management
//@{
	/**
	 * Receive an input event.
	 * The Screen can freely decide whether it wants to
	 * react to it, or not.
	 */
	virtual void receiveEvent(const EventType *evtPtr) SEC_GAME = 0;
//@}


/// @name Screen properties
//@{
    /**
     * Indicates whether the screen wants a menu.
     *
     * @return true = menu enabled, false = menu disabled
     */
    virtual bool enableMenu() const SEC_GAME;
//@}
};



#define SCREEN_PUBLIC_OPS(ClassName) \
	ClassName(const StateDescriptor *screenState) SEC_GAME; \
	virtual ~ClassName() SEC_GAME; \
 \
	void init(const Canvas *displayCanvas, const Canvas *viewportCanvas) SEC_GAME; \
 \
	void getViewportProperties(const Canvas *displayCanvas, ViewportProperties& viewport) const SEC_GAME; \
 \
	Boolean nextPeriod() SEC_GAME; \
 \
	void drawBackground(const Canvas *viewportCanvas) const SEC_GAME; \
	void drawAction(const Canvas *viewportCanvas, RectangleType *dirtyRectangle) SEC_GAME;	 \
	void drawFixedOverlays(const Canvas *displayCanvas) const SEC_GAME;	 \
	void drawDynamicOverlays(const Canvas *displayCanvas) const SEC_GAME; \
 \
	void receiveEvent(const EventType *evtPtr) SEC_GAME; 




#endif
